'''
Copyright (c) 2012-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/kairos/blob/master/LICENSE.txt
'''
from .exceptions import *
from .timeseries import *

import cql

import time
from datetime import date, datetime
from datetime import time as time_type
from decimal import Decimal
from Queue import Queue, Empty, Full
import re

# Test python3 compatibility
try:
  x = long(1)
except NameError:
  long = int
try:
  x = unicode('foo')
except NameError:
  unicode = str

TYPE_MAP = {
  str         : 'ascii',
  'str'       : 'ascii',
  'string'    : 'ascii',

  unicode     : 'text',  # works for py3 too
  'unicode'   : 'text',

  float       : 'float',
  'float'     : 'float',

  'double'    : 'double',

  int         : 'int',
  'int'       : 'int',
  'integer'   : 'int',

  long        : 'varint', # works for py3 too
  'long'      : 'varint',
  'int64'     : 'bigint',

  'decimal'   : 'decimal',

  bool        : 'boolean',
  'bool'      : 'boolean',
  'boolean'   : 'boolean',

  'text'      : 'text',
  'clob'      : 'blob',
  'blob'      : 'blob',

  'inet'      : 'inet',
}

QUOTE_TYPES = set(['ascii','text','blob'])
QUOTE_MATCH = re.compile("^'.*'$")

def scoped_connection(func):
  '''
  Decorator that gives out connections.
  '''
  def _with(series, *args, **kwargs):
    connection = None
    try:
      connection = series._connection()
      return func(series, connection, *args, **kwargs)
    finally:
      series._return( connection )
  return _with

class CassandraBackend(Timeseries):

  def __new__(cls, *args, **kwargs):
    if cls==CassandraBackend:
      ttype = kwargs.pop('type', None)
      if ttype=='series':
        return CassandraSeries.__new__(CassandraSeries, *args, **kwargs)
      elif ttype=='histogram':
        return CassandraHistogram.__new__(CassandraHistogram, *args, **kwargs)
      elif ttype=='count':
        return CassandraCount.__new__(CassandraCount, *args, **kwargs)
      elif ttype=='gauge':
        return CassandraGauge.__new__(CassandraGauge, *args, **kwargs)
      elif ttype=='set':
        return CassandraSet.__new__(CassandraSet, *args, **kwargs)
    return Timeseries.__new__(cls, *args, **kwargs)

  def __init__(self, client, **kwargs):
    '''
    Initialize the sql backend after timeseries has processed the configuration.
    '''
    # Only CQL3 is supported
    if client.cql_major_version != 3:
      raise TypeError("Only CQL3 is supported")

    vtype = kwargs.get('value_type', float)
    if vtype in TYPE_MAP:
      self._value_type = TYPE_MAP[vtype]
    else:
      raise TypeError("Unsupported type '%s'"%(vtype))

    self._table = kwargs.get('table_name', self._table)

    # copy internal variables of the connection for poor-mans pooling
    self._host = client.host
    self._port = client.port
    self._keyspace = client.keyspace
    self._cql_version = client.cql_version
    self._compression = client.compression
    self._consistency_level = client.consistency_level
    self._transport = client.transport
    self._credentials = client.credentials
    self._pool = Queue(kwargs.get('pool_size',0))
    self._pool.put( client )

    super(CassandraBackend,self).__init__(client, **kwargs)

  def _connection(self):
    '''
    Return a connection from the pool
    '''
    try:
      return self._pool.get(False)
    except Empty:
      args = [
        self._host, self._port, self._keyspace
      ]
      kwargs = {
        'user'              : None,
        'password'          : None,
        'cql_version'       : self._cql_version,
        'compression'       : self._compression,
        'consistency_level' : self._consistency_level,
        'transport'         : self._transport,
      }
      if self._credentials:
        kwargs['user'] = self._credentials['user']
        kwargs['password'] = self._credentials['password']
      return cql.connect(*args, **kwargs)

  def _return(self, connection):
    try:
      self._pool.put(connection, False)
    except Full:
      # do not return connection to the pool.
      pass

  def _insert(self, name, value, timestamp, intervals, **kwargs):
    '''
    Insert the new value.
    '''
    if self._value_type in QUOTE_TYPES and not QUOTE_MATCH.match(value):
      value = "'%s'"%(value)

    for interval,config in self._intervals.items():
      timestamps = self._normalize_timestamps(timestamp, intervals, config)
      for tstamp in timestamps:
        self._insert_data(name, value, tstamp, interval, config, **kwargs)

  @scoped_connection
  def _insert_data(self, connection, name, value, timestamp, interval, config):
    '''Helper to insert data into cql.'''
    cursor = connection.cursor()
    try:
      stmt = self._insert_stmt(name, value, timestamp, interval, config)
      if stmt:
        cursor.execute(stmt)
    finally:
      cursor.close()

  @scoped_connection
  def _get(self, connection, name, interval, config, timestamp, **kws):
    '''
    Get the interval.
    '''
    i_bucket = config['i_calc'].to_bucket(timestamp)
    fetch = kws.get('fetch')
    process_row = kws.get('process_row') or self._process_row

    rval = OrderedDict()
    if fetch:
      data = fetch( connection, self._table, name, interval, [i_bucket] )
    else:
      data = self._type_get(name, interval, i_bucket)

    if config['coarse']:
      if data:
        rval[ config['i_calc'].from_bucket(i_bucket) ] = process_row(data.values()[0][None])
      else:
        rval[ config['i_calc'].from_bucket(i_bucket) ] = self._type_no_value()
    else:
      for r_bucket,row_data in data.values()[0].items():
        rval[ config['r_calc'].from_bucket(r_bucket) ] = process_row(row_data)

    return rval

  @scoped_connection
  def _series(self, connection, name, interval, config, buckets, **kws):
    '''
    Fetch a series of buckets.
    '''
    fetch = kws.get('fetch')
    process_row = kws.get('process_row') or self._process_row

    rval = OrderedDict()

    if fetch:
      data = fetch( connection, self._table, name, interval, buckets )
    else:
      data = self._type_get(name, interval, buckets[0], buckets[-1])

    if config['coarse']:
      for i_bucket in buckets:
        i_key = config['i_calc'].from_bucket(i_bucket)
        i_data = data.get( i_bucket )
        if i_data:
          rval[ i_key ] = process_row( i_data[None] )
        else:
          rval[ i_key ] = self._type_no_value()
    else:
      if data:
        for i_bucket, i_data in data.items():
          i_key = config['i_calc'].from_bucket(i_bucket)
          rval[i_key] = OrderedDict()
          for r_bucket, r_data in i_data.items():
            r_key = config['r_calc'].from_bucket(r_bucket)
            if r_data:
              rval[i_key][r_key] = process_row(r_data)
            else:
              rval[i_key][r_key] = self._type_no_value()

    return rval

  @scoped_connection
  def delete(self, connection, name):
    cursor = connection.cursor()
    try:
      cursor.execute("DELETE FROM %s WHERE name='%s';"%(self._table,name))
    finally:
      cursor.close()

  @scoped_connection
  def delete_all(self, connection):
    cursor = connection.cursor()
    try:
      cursor.execute("TRUNCATE %s"%(self._table))
    finally:
      cursor.close()

  @scoped_connection
  def list(self, connection):
    cursor = connection.cursor()
    rval = set()

    try:
      cursor.execute('SELECT name FROM %s'%(self._table))
      for row in cursor:
        rval.add(row[0])
    finally:
      cursor.close()
    return list(rval)

  @scoped_connection
  def properties(self, connection, name):
    cursor = connection.cursor()
    rval = {}

    try:
      for interval,config in self._intervals.items():
        rval.setdefault(interval, {})

        cursor.execute('''SELECT i_time
          FROM %s
          WHERE name = '%s' AND interval = '%s'
          ORDER BY interval ASC, i_time ASC
          LIMIT 1'''%(self._table, name, interval))

        rval[interval]['first'] = config['i_calc'].from_bucket(
          cursor.fetchone()[0] )

        cursor.execute('''SELECT i_time
          FROM %s
          WHERE name = '%s' AND interval = '%s'
          ORDER BY interval DESC, i_time DESC
          LIMIT 1'''%(self._table, name, interval))
        rval[interval]['last'] = config['i_calc'].from_bucket(
          cursor.fetchone()[0] )
    finally:
      cursor.close()

    return rval

class CassandraSeries(CassandraBackend, Series):

  def __init__(self, *a, **kwargs):
    self._table = 'series'
    super(CassandraSeries,self).__init__(*a, **kwargs)

    cursor = self._client.cursor()
    # TODO: support other value types
    # TODO: use varint for [ir]_time?
    try:
      res = cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
        name text,
        interval text,
        i_time bigint,
        r_time bigint,
        value list<%s>,
        PRIMARY KEY(name, interval, i_time, r_time)
      )'''%(self._table, self._value_type))
    except cql.ProgrammingError as pe:
      if 'existing' not in str(pe):
        raise
    finally:
      cursor.close()

  def _insert_stmt(self, name, value, timestamp, interval, config):
    '''Helper to generate the insert statement.'''
    # Calculate the TTL and abort if inserting into the past
    expire, ttl = config['expire'], config['ttl'](timestamp)
    if expire and not ttl:
      return None

    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = -1

    # TODO: figure out escaping rules of CQL
    table_spec = self._table
    if ttl:
      table_spec += " USING TTL %s "%(ttl)
    stmt = '''UPDATE %s SET value = value + [%s]
      WHERE name = '%s'
      AND interval = '%s'
      AND i_time = %s
      AND r_time = %s'''%(table_spec, value, name, interval, i_time, r_time)
    return stmt

  @scoped_connection
  def _type_get(self, connection, name, interval, i_bucket, i_end=None):
    rval = OrderedDict()

    # TODO: more efficient creation of query string
    stmt = '''SELECT i_time, r_time, value
      FROM %s
      WHERE name = '%s' AND interval = '%s'
    '''%(self._table, name, interval)
    if i_end :
      stmt += ' AND i_time >= %s AND i_time <= %s'%(i_bucket, i_end)
    else:
      stmt += ' AND i_time = %s'%(i_bucket)
    stmt += ' ORDER BY interval, i_time, r_time'

    cursor = connection.cursor()
    try:
      cursor.execute(stmt)
      for row in cursor:
        i_time, r_time, value = row
        if r_time==-1:
          r_time = None
        rval.setdefault(i_time,OrderedDict())[r_time] = value
    finally:
      cursor.close()
    return rval

class CassandraHistogram(CassandraBackend, Histogram):

  def __init__(self, *a, **kwargs):
    self._table = 'histogram'
    super(CassandraHistogram,self).__init__(*a, **kwargs)

    # TODO: use varint for [ir]_time?
    # TODO: support other value types
    cursor = self._client.cursor()
    try:
      res = cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
        name text,
        interval text,
        i_time bigint,
        r_time bigint,
        value %s,
        count counter,
        PRIMARY KEY(name, interval, i_time, r_time, value)
      )'''%(self._table, self._value_type))
    except cql.ProgrammingError as pe:
      if 'existing' not in str(pe):
        raise
    finally:
      cursor.close()

  def _insert_stmt(self, name, value, timestamp, interval, config):
    '''Helper to generate the insert statement.'''
    # Calculate the TTL and abort if inserting into the past
    expire, ttl = config['expire'], config['ttl'](timestamp)
    if expire and not ttl:
      return None

    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = -1

    # TODO: figure out escaping rules of CQL
    table_spec = self._table
    if ttl:
      table_spec += " USING TTL %s "%(ttl)
    stmt = '''UPDATE %s SET count = count + 1
      WHERE name = '%s'
      AND interval = '%s'
      AND i_time = %s
      AND r_time = %s
      AND value = %s'''%(table_spec, name, interval, i_time, r_time, value)
    return stmt

  @scoped_connection
  def _type_get(self, connection, name, interval, i_bucket, i_end=None):
    rval = OrderedDict()

    # TODO: more efficient creation of query string
    stmt = '''SELECT i_time, r_time, value, count
      FROM %s
      WHERE name = '%s' AND interval = '%s'
    '''%(self._table, name, interval)
    if i_end :
      stmt += ' AND i_time >= %s AND i_time <= %s'%(i_bucket, i_end)
    else:
      stmt += ' AND i_time = %s'%(i_bucket)
    stmt += ' ORDER BY interval, i_time, r_time'

    cursor = connection.cursor()
    try:
      cursor.execute(stmt)
      for row in cursor:
        i_time, r_time, value, count = row
        if r_time==-1:
          r_time = None
        rval.setdefault(i_time,OrderedDict()).setdefault(r_time,{})[value] = count
    finally:
      cursor.close()
    return rval

class CassandraCount(CassandraBackend, Count):

  def __init__(self, *a, **kwargs):
    self._table = 'count'
    super(CassandraCount,self).__init__(*a, **kwargs)

    # TODO: use varint for [ir]_time?
    # TODO: support other value types
    cursor = self._client.cursor()
    try:
      res = cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
        name text,
        interval text,
        i_time bigint,
        r_time bigint,
        count counter,
        PRIMARY KEY(name, interval, i_time, r_time)
      )'''%(self._table))
    except cql.ProgrammingError as pe:
      if 'existing' not in str(pe):
        raise
    finally:
      cursor.close()

  def _insert_stmt(self, name, value, timestamp, interval, config):
    '''Helper to generate the insert statement.'''
    # Calculate the TTL and abort if inserting into the past
    expire, ttl = config['expire'], config['ttl'](timestamp)
    if expire and not ttl:
      return None

    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = -1

    # TODO: figure out escaping rules of CQL
    table_spec = self._table
    if ttl:
      table_spec += " USING TTL %s "%(ttl)
    stmt = '''UPDATE %s SET count = count + %s
      WHERE name = '%s'
      AND interval = '%s'
      AND i_time = %s
      AND r_time = %s'''%(table_spec, value, name, interval, i_time, r_time)
    return stmt

  @scoped_connection
  def _type_get(self, connection, name, interval, i_bucket, i_end=None):
    rval = OrderedDict()

    # TODO: more efficient creation of query string
    stmt = '''SELECT i_time, r_time, count
      FROM %s
      WHERE name = '%s' AND interval = '%s'
    '''%(self._table, name, interval)
    if i_end :
      stmt += ' AND i_time >= %s AND i_time <= %s'%(i_bucket, i_end)
    else:
      stmt += ' AND i_time = %s'%(i_bucket)
    stmt += ' ORDER BY interval, i_time, r_time'

    cursor = connection.cursor()
    try:
      cursor.execute(stmt)
      for row in cursor:
        i_time, r_time, count = row
        if r_time==-1:
          r_time = None
        rval.setdefault(i_time,OrderedDict())[r_time] = count
    finally:
      cursor.close()
    return rval

class CassandraGauge(CassandraBackend, Gauge):

  def __init__(self, *a, **kwargs):
    self._table = 'gauge'
    super(CassandraGauge,self).__init__(*a, **kwargs)

    # TODO: use varint for [ir]_time?
    # TODO: support other value types
    cursor = self._client.cursor()
    try:
      res = cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
        name text,
        interval text,
        i_time bigint,
        r_time bigint,
        value %s,
        PRIMARY KEY(name, interval, i_time, r_time)
      )'''%(self._table, self._value_type))
    except cql.ProgrammingError as pe:
      if 'existing' not in str(pe):
        raise
    finally:
      cursor.close()

  def _insert_stmt(self, name, value, timestamp, interval, config):
    '''Helper to generate the insert statement.'''
    # Calculate the TTL and abort if inserting into the past
    expire, ttl = config['expire'], config['ttl'](timestamp)
    if expire and not ttl:
      return None

    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = -1

    # TODO: figure out escaping rules of CQL
    table_spec = self._table
    if ttl:
      table_spec += " USING TTL %s "%(ttl)
    stmt = '''UPDATE %s SET value = %s
      WHERE name = '%s'
      AND interval = '%s'
      AND i_time = %s
      AND r_time = %s'''%(table_spec, value, name, interval, i_time, r_time)
    return stmt

  @scoped_connection
  def _type_get(self, connection, name, interval, i_bucket, i_end=None):
    rval = OrderedDict()

    # TODO: more efficient creation of query string
    stmt = '''SELECT i_time, r_time, value
      FROM %s
      WHERE name = '%s' AND interval = '%s'
    '''%(self._table, name, interval)
    if i_end :
      stmt += ' AND i_time >= %s AND i_time <= %s'%(i_bucket, i_end)
    else:
      stmt += ' AND i_time = %s'%(i_bucket)
    stmt += ' ORDER BY interval, i_time, r_time'

    cursor = connection.cursor()
    try:
      cursor.execute(stmt)
      for row in cursor:
        i_time, r_time, value = row
        if r_time==-1:
          r_time = None
        rval.setdefault(i_time,OrderedDict())[r_time] = value
    finally:
      cursor.close()
    return rval

class CassandraSet(CassandraBackend, Set):

  def __init__(self, *a, **kwargs):
    self._table = 'sets'
    super(CassandraSet,self).__init__(*a, **kwargs)

    # TODO: use varint for [ir]_time?
    # TODO: support other value types
    cursor = self._client.cursor()
    try:
      res = cursor.execute('''CREATE TABLE IF NOT EXISTS %s (
        name text,
        interval text,
        i_time bigint,
        r_time bigint,
        value %s,
        PRIMARY KEY(name, interval, i_time, r_time, value)
      )'''%(self._table, self._value_type))
    except cql.ProgrammingError as pe:
      if 'existing' not in str(pe):
        raise
    finally:
      cursor.close()

  def _insert_stmt(self, name, value, timestamp, interval, config):
    '''Helper to generate the insert statement.'''
    # Calculate the TTL and abort if inserting into the past
    expire, ttl = config['expire'], config['ttl'](timestamp)
    if expire and not ttl:
      return None

    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = -1

    # TODO: figure out escaping rules of CQL
    stmt = '''INSERT INTO %s (name, interval, i_time, r_time, value)
      VALUES ('%s', '%s', %s, %s, %s)'''%(self._table, name, interval, i_time, r_time, value)
    expire = config['expire']
    if ttl:
      stmt += " USING TTL %s"%(ttl)
    return stmt

  @scoped_connection
  def _type_get(self, connection, name, interval, i_bucket, i_end=None):
    rval = OrderedDict()

    # TODO: more efficient creation of query string
    stmt = '''SELECT i_time, r_time, value
      FROM %s
      WHERE name = '%s' AND interval = '%s'
    '''%(self._table, name, interval)
    if i_end :
      stmt += ' AND i_time >= %s AND i_time <= %s'%(i_bucket, i_end)
    else:
      stmt += ' AND i_time = %s'%(i_bucket)
    stmt += ' ORDER BY interval, i_time, r_time'

    cursor = connection.cursor()
    try:
      cursor.execute(stmt)
      for row in cursor:
        i_time, r_time, value = row
        if r_time==-1:
          r_time = None
        rval.setdefault(i_time,OrderedDict()).setdefault(r_time,set()).add( value )
    finally:
      cursor.close()
    return rval



'''
Copyright (c) 2012-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/kairos/blob/master/LICENSE.txt
'''
from .exceptions import *
from .timeseries import *

from sqlalchemy.types import TypeEngine
from sqlalchemy import Table, Column, BigInteger, Integer, String, Unicode, Text, LargeBinary, Float, Boolean, Time, Date, DateTime, Numeric, MetaData, UniqueConstraint
from sqlalchemy.sql import select, update, insert, distinct, asc, desc, and_, or_, not_

import time
from datetime import date, datetime
from datetime import time as time_type
from decimal import Decimal

# Test python3 compatibility
try:
  x = long(1)
except NameError:
  long = int
try:
  x = unicode('foo')
except NameError:
  unicode = str

TYPE_MAP = {
  str         : String,
  'str'       : String,
  'string'    : String,

  unicode     : Unicode,  # works for py3 too
  'unicode'   : Unicode,

  float       : Float,
  'float'     : Float,

  int         : Integer,
  'int'       : Integer,
  'integer'   : Integer,

  long        : BigInteger, # works for py3 too
  'long'      : BigInteger,
  'int64'     : BigInteger,

  bool        : Boolean,
  'bool'      : Boolean,
  'boolean'   : Boolean,

  date        : Date,
  'date'      : Date,
  datetime    : DateTime,
  'datetime'  : DateTime,
  time_type   : Time,
  'time'      : Time,

  Decimal     : Numeric,
  'decimal'   : Numeric,

  'text'      : Text,
  'clob'      : Text,
  'blob'      : LargeBinary,
}

class SqlBackend(Timeseries):

  def __new__(cls, *args, **kwargs):
    if cls==SqlBackend:
      ttype = kwargs.pop('type', None)
      if ttype=='series':
        return SqlSeries.__new__(SqlSeries, *args, **kwargs)
      elif ttype=='histogram':
        return SqlHistogram.__new__(SqlHistogram, *args, **kwargs)
      elif ttype=='count':
        return SqlCount.__new__(SqlCount, *args, **kwargs)
      elif ttype=='gauge':
        return SqlGauge.__new__(SqlGauge, *args, **kwargs)
    return Timeseries.__new__(cls, *args, **kwargs)

  def __init__(self, client, **kwargs):
    '''
    Initialize the sql backend after timeseries has processed the configuration.
    '''
    self._metadata = MetaData()
    self._str_length = kwargs.get('string_length',255)
    self._txt_length = kwargs.get('text_length', 32*1024)

    vtype = kwargs.get('value_type', float)
    if vtype in TYPE_MAP:
      self._value_type = TYPE_MAP[vtype]
      if self._value_type == String:
        self._value_type = String(self._str_length)
      elif self._value_type == Text:
        self._value_type = Text(self._txt_length)
      elif self._value_type == LargeBinary:
        self._value_type = LargeBinary(self._txt_length)

    elif issubclass(vtype, TypeEngine):
      if vtype == String:
        self._value_type = String(self._str_length)
      elif vtype == Text:
        self._value_type = Text(self._txt_length)
      elif vtype == LargeBinary:
        self._value_type = LargeBinary(self._txt_length)

    elif isinstance(vtype, TypeEngine):
      self._value_type = vtype

    else:
      raise ValueError("Unsupported type '%s'"%(vtype))

    self._table_name = kwargs.get('table_name', self._table_name)

    super(SqlBackend,self).__init__(client, **kwargs)

  def list(self):
    connection = self._client.connect()
    rval = set()
    stmt = select([distinct(self._table.c.name)])

    for row in connection.execute(stmt):
      rval.add(row['name'])
    return list(rval)

  def properties(self, name):
    connection = self._client.connect()
    rval = {}

    for interval,config in self._intervals.items():
      rval.setdefault(interval, {})

      stmt = select([self._table.c.i_time]).where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval
        )
      ).order_by( asc(self._table.c.i_time) ).limit(1)
      rval[interval]['first'] = config['i_calc'].from_bucket(
        connection.execute(stmt).first()['i_time'] )

      stmt = select([self._table.c.i_time]).where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval
        )
      ).order_by( desc(self._table.c.i_time) ).limit(1)
      rval[interval]['last'] = config['i_calc'].from_bucket(
        connection.execute(stmt).first()['i_time'] )

    return rval

  def expire(self, name):
    '''
    Expire all the data.
    '''
    for interval,config in self._intervals.items():
      if config['expire']:
        # Because we're storing the bucket time, expiry has the same
        # "skew" as whatever the buckets are.
        expire_from = config['i_calc'].to_bucket(time.time() - config['expire'])
        conn = self._client.connect()

        conn.execute( self._table.delete().where(
          and_(
            self._table.c.name==name,
            self._table.c.interval==interval,
            self._table.c.i_time<=expire_from
          )
        ))

  def _insert(self, name, value, timestamp, intervals, **kwargs):
    '''
    Insert the new value.
    '''
    for interval,config in self._intervals.items():
      timestamps = self._normalize_timestamps(timestamp, intervals, config)
      for tstamp in timestamps:
        self._insert_data(name, value, tstamp, interval, config, **kwargs)

  def _get(self, name, interval, config, timestamp, **kws):
    '''
    Get the interval.
    '''
    i_bucket = config['i_calc'].to_bucket(timestamp)
    fetch = kws.get('fetch')
    process_row = kws.get('process_row') or self._process_row

    rval = OrderedDict()
    if fetch:
      data = fetch( self._client.connect(), self._table, name, interval, i_bucket )
    else:
      data = self._type_get(name, interval, i_bucket)

    if config['coarse']:
      if data:
        rval[ config['i_calc'].from_bucket(i_bucket) ] = process_row(data.values()[0][None])
      else:
        rval[ config['i_calc'].from_bucket(i_bucket) ] = self._type_no_value()
    else:
      for r_bucket,row_data in data.values()[0].items():
        rval[ config['r_calc'].from_bucket(r_bucket) ] = process_row(row_data)

    return rval

  def _series(self, name, interval, config, buckets, **kws):
    '''
    Fetch a series of buckets.
    '''
    fetch = kws.get('fetch')
    process_row = kws.get('process_row') or self._process_row

    rval = OrderedDict()

    if fetch:
      data = fetch( self._client.connect(), self._table, name, interval, buckets[0], buckets[-1] )
    else:
      data = self._type_get(name, interval, buckets[0], buckets[-1])

    if config['coarse']:
      for i_bucket in buckets:
        i_key = config['i_calc'].from_bucket(i_bucket)
        i_data = data.get( i_bucket )
        if i_data:
          rval[ i_key ] = process_row( i_data[None] )
        else:
          rval[ i_key ] = self._type_no_value()
    else:
      if data:
        for i_bucket, i_data in data.items():
          i_key = config['i_calc'].from_bucket(i_bucket)
          rval[i_key] = OrderedDict()
          for r_bucket, r_data in i_data.items():
            r_key = config['r_calc'].from_bucket(r_bucket)
            if r_data:
              rval[i_key][r_key] = process_row(r_data)
            else:
              rval[i_key][r_key] = self._type_no_value()

    return rval

  def delete(self, name):
    '''
    Delete time series by name across all intervals. Returns the number of
    records deleted.
    '''
    conn = self._client.connect()
    conn.execute( self._table.delete().where(self._table.c.name==name) )

class SqlSeries(SqlBackend, Series):

  def __init__(self, *a, **kwargs):
    # TODO: define indices
    # TODO: optionally create separate tables for each interval, like mongo?
    self._table_name = 'series'
    super(SqlSeries,self).__init__(*a, **kwargs)
    self._table = Table(self._table_name, self._metadata,
      Column('name', String(self._str_length), nullable=False),      # stat name
      Column('interval', String(self._str_length), nullable=False),  # interval name
      Column('insert_time', Float, nullable=False),     # to preserve order
      Column('i_time', Integer, nullable=False),        # interval timestamp
      Column('r_time', Integer, nullable=True),         # resolution timestamp
      Column('value', self._value_type, nullable=False)            # datas
    )
    self._metadata.create_all(self._client)

  def _insert_data(self, name, value, timestamp, interval, config, **kwargs):
    '''Helper to insert data into sql.'''
    kwargs = {
      'name'        : name,
      'interval'    : interval,
      'insert_time' : time.time(),
      'i_time'      : config['i_calc'].to_bucket(timestamp),
      'value'       : value
    }
    if not config['coarse']:
      kwargs['r_time'] = config['r_calc'].to_bucket(timestamp)
    stmt = self._table.insert().values(**kwargs)
    conn = self._client.connect()
    result = conn.execute(stmt)

  def _type_get(self, name, interval, i_bucket, i_end=None):
    connection = self._client.connect()
    rval = OrderedDict()
    stmt = self._table.select()

    if i_end:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time>=i_bucket,
          self._table.c.i_time<=i_end,
        )
      )
    else:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time==i_bucket,
        )
      )
    stmt = stmt.order_by( self._table.c.r_time, self._table.c.insert_time )

    for row in connection.execute(stmt):
      rval.setdefault(row['i_time'],OrderedDict()).setdefault(row['r_time'],[]).append( row['value'] )
    return rval

class SqlHistogram(SqlBackend, Histogram):

  def __init__(self, *a, **kwargs):
    # TODO: define indices
    # TODO: optionally create separate tables for each interval, like mongo?
    self._table_name = 'histogram'
    super(SqlHistogram,self).__init__(*a, **kwargs)
    self._table = Table(self._table_name, self._metadata,
      Column('name', String(self._str_length), nullable=False),      # stat name
      Column('interval', String(self._str_length), nullable=False),  # interval name
      Column('i_time', Integer, nullable=False),        # interval timestamp
      Column('r_time', Integer, nullable=True),         # resolution timestamp
      Column('value', self._value_type, nullable=False),           # histogram keys
      Column('count', Integer, nullable=False),         # key counts

      # Use a constraint for transaction-less insert vs update
      UniqueConstraint('name', 'interval', 'i_time', 'r_time', 'value', name='unique_value')
    )
    self._metadata.create_all(self._client)

  def _insert_data(self, name, value, timestamp, interval, config, **kwargs):
    '''Helper to insert data into sql.'''
    conn = self._client.connect()
    if not self._update_data(name, value, timestamp, interval, config, conn):
      try:
        kwargs = {
          'name'        : name,
          'interval'    : interval,
          'i_time'      : config['i_calc'].to_bucket(timestamp),
          'value'       : value,
          'count'       : 1
        }
        if not config['coarse']:
          kwargs['r_time'] = config['r_calc'].to_bucket(timestamp)
        stmt = self._table.insert().values(**kwargs)
        result = conn.execute(stmt)
      except:
        # TODO: only catch IntegrityError
        if not self._update_data(name, value, timestamp, interval, config, conn):
          raise

  def _update_data(self, name, value, timestamp, interval, config, conn):
    '''Support function for insert. Should be called within a transaction'''
    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = None
    stmt = self._table.update().where(
      and_(
        self._table.c.name==name,
        self._table.c.interval==interval,
        self._table.c.i_time==i_time,
        self._table.c.r_time==r_time,
        self._table.c.value==value)
    ).values({self._table.c.count: self._table.c.count + 1})
    rval = conn.execute( stmt )
    return rval.rowcount

  def _type_get(self, name, interval, i_bucket, i_end=None):
    connection = self._client.connect()
    rval = OrderedDict()
    stmt = self._table.select()

    if i_end:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time>=i_bucket,
          self._table.c.i_time<=i_end,
        )
      )
    else:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time==i_bucket,
        )
      )
    stmt = stmt.order_by( self._table.c.r_time )

    for row in connection.execute(stmt):
      rval.setdefault(row['i_time'],OrderedDict()).setdefault(row['r_time'],{})[row['value']] = row['count']
    return rval

class SqlCount(SqlBackend, Count):

  def __init__(self, *a, **kwargs):
    # TODO: define indices
    # TODO: optionally create separate tables for each interval, like mongo?
    self._table_name = 'count'
    super(SqlCount,self).__init__(*a, **kwargs)
    self._table = Table(self._table_name, self._metadata,
      Column('name', String(self._str_length), nullable=False),      # stat name
      Column('interval', String(self._str_length), nullable=False),  # interval name
      Column('i_time', Integer, nullable=False),        # interval timestamp
      Column('r_time', Integer, nullable=True),         # resolution timestamp
      Column('count', Integer, nullable=False),         # key counts

      # Use a constraint for transaction-less insert vs update
      UniqueConstraint('name', 'interval', 'i_time', 'r_time', name='unique_count')
    )
    self._metadata.create_all(self._client)

  def _insert_data(self, name, value, timestamp, interval, config, **kwargs):
    '''Helper to insert data into sql.'''
    conn = self._client.connect()
    if not self._update_data(name, value, timestamp, interval, config, conn):
      try:
        kwargs = {
          'name'        : name,
          'interval'    : interval,
          'i_time'      : config['i_calc'].to_bucket(timestamp),
          'count'       : value
        }
        if not config['coarse']:
          kwargs['r_time'] = config['r_calc'].to_bucket(timestamp)
        stmt = self._table.insert().values(**kwargs)
        result = conn.execute(stmt)
      except:
        # TODO: only catch IntegrityError
        if not self._update_data(name, value, timestamp, interval, config, conn):
          raise

  def _update_data(self, name, value, timestamp, interval, config, conn):
    '''Support function for insert. Should be called within a transaction'''
    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = None
    stmt = self._table.update().where(
      and_(
        self._table.c.name==name,
        self._table.c.interval==interval,
        self._table.c.i_time==i_time,
        self._table.c.r_time==r_time)
    ).values({self._table.c.count: self._table.c.count + value})
    rval = conn.execute( stmt )
    return rval.rowcount

  def _type_get(self, name, interval, i_bucket, i_end=None):
    connection = self._client.connect()
    rval = OrderedDict()
    stmt = self._table.select()

    if i_end:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time>=i_bucket,
          self._table.c.i_time<=i_end,
        )
      )
    else:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time==i_bucket,
        )
      )
    stmt = stmt.order_by( self._table.c.r_time )

    for row in connection.execute(stmt):
      rval.setdefault(row['i_time'],OrderedDict())[row['r_time']] = row['count']
    return rval

class SqlGauge(SqlBackend, Gauge):

  def __init__(self, *a, **kwargs):
    # TODO: define indices
    # TODO: optionally create separate tables for each interval, like mongo?
    self._table_name = 'gauge'
    super(SqlGauge,self).__init__(*a, **kwargs)
    self._table = Table(self._table_name, self._metadata,
      Column('name', String(self._str_length), nullable=False),      # stat name
      Column('interval', String(self._str_length), nullable=False),  # interval name
      Column('i_time', Integer, nullable=False),        # interval timestamp
      Column('r_time', Integer, nullable=True),         # resolution timestamp
      Column('value', self._value_type, nullable=False),           # key counts

      # Use a constraint for transaction-less insert vs update
      UniqueConstraint('name', 'interval', 'i_time', 'r_time', name='unique_count')
    )
    self._metadata.create_all(self._client)

  def _insert_data(self, name, value, timestamp, interval, config, **kwargs):
    '''Helper to insert data into sql.'''
    conn = self._client.connect()
    if not self._update_data(name, value, timestamp, interval, config, conn):
      try:
        kwargs = {
          'name'        : name,
          'interval'    : interval,
          'i_time'      : config['i_calc'].to_bucket(timestamp),
          'value'       : value
        }
        if not config['coarse']:
          kwargs['r_time'] = config['r_calc'].to_bucket(timestamp)
        stmt = self._table.insert().values(**kwargs)
        result = conn.execute(stmt)
      except:
        # TODO: only catch IntegrityError
        if not self._update_data(name, value, timestamp, interval, config, conn):
          raise

  def _update_data(self, name, value, timestamp, interval, config, conn):
    '''Support function for insert. Should be called within a transaction'''
    i_time = config['i_calc'].to_bucket(timestamp)
    if not config['coarse']:
      r_time = config['r_calc'].to_bucket(timestamp)
    else:
      r_time = None
    stmt = self._table.update().where(
      and_(
        self._table.c.name==name,
        self._table.c.interval==interval,
        self._table.c.i_time==i_time,
        self._table.c.r_time==r_time)
    ).values({self._table.c.value: value})
    rval = conn.execute( stmt )
    return rval.rowcount

  def _type_get(self, name, interval, i_bucket, i_end=None):
    connection = self._client.connect()
    rval = OrderedDict()
    stmt = self._table.select()

    if i_end:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time>=i_bucket,
          self._table.c.i_time<=i_end,
        )
      )
    else:
      stmt = stmt.where(
        and_(
          self._table.c.name==name,
          self._table.c.interval==interval,
          self._table.c.i_time==i_bucket,
        )
      )
    stmt = stmt.order_by( self._table.c.r_time )

    for row in connection.execute(stmt):
      rval.setdefault(row['i_time'],OrderedDict())[row['r_time']] = row['value']
    return rval



'''
Copyright (c) 2012-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/kairos/blob/master/LICENSE.txt
'''
from __future__ import absolute_import
__version__ = "0.9.0"

from .timeseries import Timeseries
from .exceptions import *
